module Philosophers where

-- The dining philosophers, using Software Transactional Memory.
-- Uses code from http://rosettacode.org/wiki/Dining_philosophers#Haskell 
 
import Control.Monad
import Control.Concurrent
import Control.Concurrent.STM
import System.Random
import System.Environment(getArgs)
 
-- A fork state:  either on the table (down), or taken. 
data FS = Down | Taken deriving Eq

-- A fork has a state
newtype Fork = Fork (TVar FS)

-- How long to think or eat.
delay :: IO ()
delay = do 
   d<- randomRIO(0,1000)
   threadDelay (d*1000) --- between 0.1 and 1 sec

newFork :: IO Fork
newFork = atomically $ do 
  f<- newTVar Down
  return $ Fork f
               
-- Take a fork
takeFork :: Fork -> STM ()
takeFork (Fork f) = do 
  s<- readTVar f
  when (s == Taken) retry 
  writeTVar f Taken 
 
-- Release a fork
releaseFork :: Fork -> STM ()
releaseFork (Fork f) = writeTVar f Down
 
philosopher :: String -> Fork-> Fork -> IO ()
philosopher name left right = forever $ do
  putStrLn $ name ++ " is thinking."
  delay
  putStrLn $ name ++ " is hungry."
  -- Take the forks 
  atomically $ do  {takeFork left; takeFork right}
  putStrLn $ name ++ " is eating."
  delay 
  putStrLn $ name ++ " is done eating. Going back to thinking." 
  -- Put down the forks
  atomically $ do {releaseFork left; releaseFork right}
 

run :: Int -> IO ()
run n = do
  forks <- replicateM n newFork
  forM_ [0..n-1] $ \i-> 
     forkIO (philosopher ("Philosopher #"++ show i) 
                         (forks !! i)
                         (forks !! ((i+1) `mod` n))) 
  getLine
  return ()

main :: IO ()
main = do
  args <- getArgs
  case args of []  -> run 5
               x:_ -> run $ read x
